/// HOW TO USE THIS SCRIPT:

/// CALLING DIRECTLY (STANDALONE MODE):
/// From a Windows command prompt, enter the command:
///   cscript StartupObs-PWI4-standalone.js
///
/// CALLING FROM ACP SCHEDULER:
///
/// 1. Move the following call to InitializePwi4() to someplace in the main() function in StartupObs.js
InitializePwi4();

/// 2. Move all of the following code to the end of the StartupObs.js file

/// 3. Edit the sequence of startup steps below (e.g. ConnectToMount(), TurnOnMountMotors())
///    to enable or disable features as needed.

/// BEGIN code for initializing PlaneWave hardware
function InitializePwi4()
{
    var PWI4_PORT = 8220;  // CHANGE THIS to the port that PWI4 is configured to use for its server
    var MOTOR_INITIALIZATION_SECONDS = 5; // Wait this many seconds for the motors to start up when first enabling after a power cycle.
                                          // May need to be longer for CDK700 motors.

    // END CONFIGURATION
    
    // BEGIN PWI Startup Script.
    // Comment or uncomment the following as needed.
    // Any line beginning with 2 forward slashes "//" will be SKIPPED.

    LaunchPwi4();

    ConnectToMount();
    TurnOnMountMotors();
    MountFindHome();          // L-Series mounts only!

    //PWI3RotatorFindHome();  // For IRF90 rotators connected via PWI3. Requires PWI version 3.4.4 or greater
    //PWI3FansNormal();       // For EFA-based fans controlled via PWI3. Requires PWI version 3.5.0 beta2 or greater
    //PWI3FansOff();          // For EFA-based fans controlled via PWI3. Requires PWI version 3.5.0 beta2 or greater

    ConnectToFocuser();       // Typically used with Series 5 hardware; connects to focuser, rotator, and fans
    TurnOnFocuserMotor();     // Series 5 or PW1000
    TurnOnRotatorMotor();     // Series 5 or PW1000
    TurnOnFans();             // Series 5 or PW1000
    //MoveM3ToPort(1);        // PW1000 only!
    //EnableDerotation();     // PW1000 only! 

    //OpenMirrorCover();      // Only if an automated mirror cover is installed

    WriteLog("Startup script completed successfully");

    // END PWI Startup Script


    // BEGIN high-level functions.

    function LaunchPwi4()
    {
        var pwi4Path = "C:\\Program Files (x86)\\PlaneWave Instruments\\PlaneWave Interface 4\\PWI4.exe";

        RunProg(pwi4Path);
        Sleep(5000); // Wait a bit for the program to finish launching
    }

    function OpenMirrorCover()
    {
        var pwshutterPath = "C:\\Program Files (x86)\\PlaneWave Instruments\\PlaneWave Shutter Control\\pwshutter.exe";

        WriteLog("Connecting to mirror cover");
        try { 
            var exitCode = RunProg(pwshutterPath + " connect", true); 
            if (exitCode != 0) {
                throw new Error(0x80040001, "Failed to connect to PWShutter");
            }
        } 
        catch (ex) { 
            WriteLog("ex1 " + ex.message); 
            throw ex; 
        }
        Sleep(5000);

        WriteLog("Opening mirror cover");
        try { 
            var exitCode = RunProg(pwshutterPath + " open", true); 
            if (exitCode != 0) {
                throw new Error(0x80040001, "Failed to open mirror covers using PWShutter");
            }
        } 
        catch (ex) { 
            WriteLog("ex2 " + ex.message); 
            throw ex; 
        }
    }

    function ConnectToMount()
    {
        // Assuming PWI4 is already running, send the command
        // to connect PWI4 to the mount

        WriteLog("Connecting to mount...");
        PwiRequest("/mount/connect");

        WriteLog("Checking mount connection...");
        var response = PwiRequest("/status");

        if (ResponseHasValue(response, "mount.is_connected", "true")) {
            WriteLog("CONNECTED!!");
        }
        else {
            throw new Error(0x80040001, "===== Failed to connect to mount");
        }
    }

    function ConnectToFocuser()
    {
        // Assuming PWI4 is already running, send the command
        // to connect PWI4 to the focuser. For devices where the
        // focuser and rotator share a single connection (such
        // as most Series 5 hardware), this will also result in the
        // rotator being connected

        WriteLog("Connecting to focuser...");
        PwiRequest("/focuser/connect");

        // Wait a moment for connection
        Sleep(3000);

        WriteLog("Checking focuser connection...");
        var response = PwiRequest("/status");

        if (ResponseHasValue(response, "focuser.is_connected", "true")) {
            WriteLog("CONNECTED!!");
        }
        else {
            throw new Error(0x80040001, "===== Failed to connect to focuser");
        }
    }
    
    function TurnOnFans()
    {
        WriteLog("Turning on fans...");
        PwiRequest("/fans/on");
    }

    function MountFindHome()
    {
        WriteLog("Finding mount home position (wait 45 seconds)...");
        PwiRequest("/mount/find_home");

        // TODO: Wait until homing is complete, or detect errors.
        // For now we'll just wait awhile and assume it's complete.
        Sleep(45000);
    }

    function PWI3RotatorFindHome()
    {
	// NOTE: This requires PWI3 version 3.4.4 or greater
        WriteLog("Finding rotator home position (wait 180 seconds)...");
        Pwi3Request("?device=rotator&cmd=findhome");

        // TODO: Wait until homing is complete, or detect errors.
        // For now we'll just wait awhile and assume it's complete.
	// It could take up to 3 minute sfor homing to complete.
        Sleep(180000);
    }

    function PWI3FansOff()
    {
	// NOTE: This requires PWI3 version 3.5.0 beta2 or greater
        WriteLog("Forcing PWI3 fans off...");
        Pwi3Request("?device=fans&mode=off");
    }

    function PWI3FansNormal()
    {
	// NOTE: This requires PWI3 version 3.5.0 beta2 or greater
        WriteLog("Setting PWI3 fans to configured behavior");
        Pwi3Request("?device=fans&mode=normal");
    }


    function TurnOnMountMotors()
    {
        WriteLog("Enabling azimuth motor...");
        PwiRequest("/mount/enable?axis=0");

        WriteLog("Enabling altitude motor...");
        PwiRequest("/mount/enable?axis=1");

        // In some cases (e.g. the CDK700) this could take up to 45 seconds to
        // complete the first time the motors are enabled after power-up
        WriteLog("Waiting " + MOTOR_INITIALIZATION_SECONDS + " seconds for motors to initialize...");
        Sleep(1000 * MOTOR_INITIALIZATION_SECONDS);

        WriteLog("Checking motors...");
        var response = PwiRequest("/status");

        azmEnabled = ResponseHasValue(response, "mount.axis0.is_enabled", "true");
        altEnabled = ResponseHasValue(response, "mount.axis1.is_enabled", "true");

        WriteLog("Azm enabled: " + azmEnabled);
        WriteLog("Alt enabled: " + altEnabled);

        if (azmEnabled && altEnabled) {
            WriteLog("Mount motors energized successfully");
        }
        else {
            throw new Error(0x80040001, "===== Failed to energize mount motors");
        }
    }

    function TurnOnFocuserMotor()
    {
        // TODO: Enable focuser1 and focuser2 explicitly. Currently enables whichever one M3 is pointing at
        WriteLog("Enabling focuser motor...");
        PwiRequest("/focuser/enable");

        WriteLog("Checking motor...");
        var response = PwiRequest("/status");
        focEnabled = ResponseHasValue(response, "focuser.is_enabled", "true");

        WriteLog("Focuser enabled: " + focEnabled);

        if (focEnabled) {
            WriteLog("Focuser energized successfully");
        }
        else {
            throw new Error(0x80040001, "===== Failed to energize focuser motors");
        }
    }

    function TurnOnRotatorMotor()
    {
        // TODO: Enable rotator1 and rotator2 explicitly. Currently enables whichever one M3 is pointing at
        WriteLog("Enabling rotator motor...");
        PwiRequest("/rotator/enable");

        Sleep(2000); // Wait a bit for motor to enable

        WriteLog("Checking motor...");
        var response = PwiRequest("/status");
        rotEnabled = ResponseHasValue(response, "rotator.is_enabled", "true");

        WriteLog("Rotator enabled: " + rotEnabled);

        if (rotEnabled) {
            WriteLog("Rotator energized successfully");
        }
        else {
            throw new Error(0x80040001, "===== Failed to energize rotator motors");
        }
    }

    function MoveM3ToPort(port)
    {
        WriteLog("Moving M3 to port " + port);
        PwiRequest("/m3/goto?port=" + port);

        for (var i = 0; i < 20; i++) {
            var response = PwiRequest("/status");
            var isAtTargetPort = ResponseHasValue(response, "m3.port", port);

            if (isAtTargetPort) {
                WriteLog("Arrived at target port");
                return;
            }
            Sleep(1000);
        }

        throw new Error(0x80040001, "===== Failed to move M3 to target port " + port);
    }

    function EnableDerotation()
    {
        // Turn on de-rotation checkbox
        WriteLog("Enabling Alt-Az de-rotation (TODO: Not yet implemented)");
        //PwiRequest("/rotator/derotatestart");
    }

    // END HIGH LEVEL FUNCTIONS
 
    // BEGIN UTILITY FUNCTIONS used for talking to PWI

    // Issue a request to the PWI4 HTTP server
    function PwiRequest(command) {
        var http = new ActiveXObject("MSXML2.ServerXMLHTTP");
        var url = "http://localhost:" + PWI4_PORT + command; 
        
        try {
            http.open("GET", url, false);
            http.setRequestHeader("Cache-Control", "max-age=0"); // Avoid getting a previous cached response back
            http.send();
        }
        catch (ex) {
            throw new Error(0x80040001, "PWI request failed. Is PWI running? Is the server enabled? Is the correct port being used?");
        }

        var responseText = http.responseText;
        return http.responseText;
    }

    // Issue a request to the PWI3 HTTP server
    function Pwi3Request(command) {
        var http = new ActiveXObject("MSXML2.ServerXMLHTTP");
        var url = "http://localhost:" + 55555 + command; 
        
        try {
            http.open("GET", url, false);
            http.setRequestHeader("Cache-Control", "max-age=0"); // Avoid getting a previous cached response back
            http.send();
        }
        catch (ex) {
            throw new Error(0x80040001, "PWI3 request failed. Is PWI3 running?");
        }

        var responseText = http.responseText;
        return http.responseText;
    }

    function ResponseHasValue(response, keyword, desiredValue)
    {
        // Search keyword=value text in response for:
        var regex = "^"                    // The beginning of a line
                  + keyword                // Followed by the keyword
                  + "="                    // Followed by an "equals" sign
                  + desiredValue           // Followed by the desired value
                  + "$"                    // Followed by the end of a line
                  ;

        var multilineFlag = "m"; // Allow "^" and "$" to work in a multiline string

        return new RegExp(regex, multilineFlag).test(response);
    }

    function WriteLog(text) {
        if (typeof(WScript) === "object") {
            // Running via cscript
            WScript.Echo(text); 
        }
        else if (typeof(Console) === "object" && typeof(Console.PrintLine) !== "undefined") {
            // Running via ACP
            Console.PrintLine(text);
        }
    }

    function Sleep(msec) {
        if (typeof(WScript) === "object") {
            // Running via cscript
            WScript.Sleep(msec); 
        }
        else if (typeof(Util) === "object" && typeof(Util.WaitForMilliseconds) !== "undefined") {
            // Running via ACP
            Util.WaitForMilliseconds(msec);
        }
    }

    function RunProg(command, waitForReturn) {
        var wshRunning = 0;
        var wshFinished = 1;
        var wshFailed = 2;

        var wshShell = new ActiveXObject("WScript.Shell");

        WriteLog("Running " + command);
        var process = wshShell.Exec(command);

        if (! waitForReturn) {
            return; // Don't wait for program to exit
        }

        while (process.Status == wshRunning) {
            Sleep(100);
        }

        WriteLog("Exited with code " + process.ExitCode);

        if (process.Status == wshFailed) {
            WriteLog("FAILED to run");
        }

        return process.ExitCode;
    }

    // END UTILITY FUNCTIONS

}
/// END code for initializing PlaneWave hardware

